/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package de.bsvrz.vew.sim.sim.datengenerator;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.DataState;
import de.bsvrz.dav.daf.main.OneSubscriptionPerSendData;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.AttributeGroupUsage;
import de.bsvrz.dav.daf.main.config.AttributeGroupUsage.Usage;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.vew.sim.sim.simulation.OnlineSimulation;
/**
 * * Simulations-Datengenerator fr Online-Simulationen 
 *
 * @version $Revision: 1.4 $ / $Date: 2008/12/03 14:00:32 $ / ($Author: Sans $)
 *
 * @author Dambach-Werke GmbH
 * @author Stefan Sans
 *
 */
public class DatenGeneratorOnlineSimulation extends DatenGeneratorSimulation 
  implements ClientReceiverInterface
{
  /**
   * Debug-Logger fr Logging-Ausgaben
   */
  private final static Debug _debug = Debug.getLogger();
 
  /**
   * Ist der Datengenerator gestartet?
   */
  private boolean _isGestartet;
  
  /**
   * Alle Empfangsanmeldungen 
   */
  private Map<String, DatenSpezifikation> _empfangsAnmeldungen;
  
  /**
   * Alle Sendeanmeldungen 
   */
  private Map<String, DatenSpezifikation> _sendeAnmeldungen;

  /**
   * Erzeuge neue Instanz des Datengenerators
   * @param simulation Die Online-Simulation, fr die der Datengenerator 
   * erzeugt werden soll
   */
  public DatenGeneratorOnlineSimulation(OnlineSimulation simulation)
  {
    super(simulation);
  }
  
  @Override
  protected OnlineSimulation getSimulation()
  {
    return (OnlineSimulation)super.getSimulation();
  }
  
  @Override
  public boolean isGestartet()
  {
    return _isGestartet;
  }
  
  @Override
  protected void init()
  {
    Collection<DatenSpezifikation> eingangsDaten = analysiereEingangsDaten();
    _empfangsAnmeldungen = new HashMap<String, DatenSpezifikation>();
    _debug.finer("Starte " + eingangsDaten.size() + 
        " Empfangsanmeldungen fr Simulation " + getSimulation());
     
    for (DatenSpezifikation ds: eingangsDaten)
    {
      try
      {
        getConnection().subscribeReceiver(this, ds.getSystemObjekte(), 
            ds, ReceiveOptions.normal(), ReceiverRole.receiver());
        String key = DatenSpezifikation.erzeugeSchluessel(ds.getAttributeGroup(), 
            ds.getAspect());
        _empfangsAnmeldungen.put(key, ds);
        
        _debug.finest(ds + " erfolgreich angemeldet auf " + ds.getSystemObjekte().size() + 
            " Objekte fr Simulation " + getSimulation());
      }
      catch (Exception e)
      {
        _debug.error("Simulation " + getSimulation() + ": Empfangsanmeldung mit " + 
            ds + " fehlgeschlagen - fr " + ds.getSystemObjekte().size() + " Objekte", e);
      }
    }
    eingangsDaten.clear();
    _debug.fine("Fr Simulation " + getSimulation() + " Online-Datengenerator initilalisiert ");
  } 
  
  
  @Override
  public void start()
  {
    init();
    _isGestartet = true;
  }
  
  @Override
  public void stop()
  {
    _isGestartet = false;
    
    abmeldenSenden();
    _sendeAnmeldungen = null;
    
    abmeldenEmpfangen();
    _empfangsAnmeldungen = null;
    
    System.gc();
  }
   
  /**
   * Abmeldung aller Empfangsanmeldungen
   */
  private void abmeldenEmpfangen()
  {
    _debug.fine("Abmelden von " + _empfangsAnmeldungen.size() +
        " Empfangsanmeldungen fr Simulation " + getSimulation());
    for (DatenSpezifikation ds: _empfangsAnmeldungen.values())
    {
      try
      {
        getConnection().unsubscribeReceiver(this, ds.getSystemObjekte(), ds);
      }
      catch (Exception e)
      {
        _debug.error("Fehler beim abmelden von "  + ds + " fr Simulation " + 
            getSimulation(), e);      
      }  
    }
    _empfangsAnmeldungen.clear();
  }
  
  /**
   * Abmeldung aller Sendeanmeldungen
   */
  private void abmeldenSenden()
  {
    if (_sendeAnmeldungen == null)
    {
      return;
    }
    _debug.fine("Abmelden von " + _sendeAnmeldungen.size() +
        " SendeAnmeldungen fr Simulation " + getSimulation());
    for (DatenSpezifikation ds: _sendeAnmeldungen.values())
    {
      try
      {
        getConnection().unsubscribeSender(this, ds.getSystemObjekte(), ds);
      }
      catch (Exception e)
      {
        _debug.error("Fehler beim abmelden von "  + ds + " fr Simulation " + 
            getSimulation(), e);      
      }  
    }  
    _sendeAnmeldungen.clear();
  }

  /* (non-Javadoc)
   * @see stauma.dav.clientside.ClientReceiverInterface#update(stauma.dav.clientside.ResultData[])
   */
  public void update(ResultData[] results)
  {
    Collection<ResultData> sendePuffer = new ArrayList<ResultData>();
    
    for (ResultData resultData: results)
    {
      if (isGestartet() && resultData.getDataState() == DataState.DATA)
      {
        AttributeGroup atg = resultData.getDataDescription().getAttributeGroup();
        Aspect asp = resultData.getDataDescription().getAspect();
        String key = DatenSpezifikation.erzeugeSchluessel(atg, asp);
        DatenSpezifikation sa;
        
        synchronized(this) // Synchronisierung fr den Zugriff auf Sendeanmeldungen
        {
          if (_sendeAnmeldungen == null)
          {
            _sendeAnmeldungen = new HashMap<String, DatenSpezifikation>();
          }
          sa = _sendeAnmeldungen.get(key);
          boolean abmeldenErlaubt = true;
          if (sa == null) // atg/asp-Kombination ist noch nicht zum Senden angemeldet
          {
            DatenSpezifikation ea = _empfangsAnmeldungen.get(key);
            if (ea != null)
            {  
              Aspect aspekt = asp;
//              SenderRole role = SenderRole.sender();
              SenderRole role = SenderRole.source();
              
              if (atg.isParameter() && asp.getPid().equals("asp.parameterSoll"))
              {
                aspekt = getConnection().getDataModel().getAspect("asp.parameterVorgabe");
                role = SenderRole.sender();
              }
              else
              {
                AttributeGroupUsage atgv = atg.getAttributeGroupUsage(asp);
                Usage benutzung = atgv.getUsage();
                
//                if (benutzung == Usage.OnlineDataAsSourceReceiver)
//                {
//                  role = SenderRole.source();
//                }  
                if (benutzung == Usage.OnlineDataAsSenderDrain)
                {
                  role = SenderRole.sender();
                }
              }
              sa = new DatenSpezifikation(atg, aspekt, getSimulation().getSimulationsVariante());
              sa.getSystemObjekte().addAll(ea.getSystemObjekte());
                            
              try
              {
                getConnection().subscribeSender(this, sa.getSystemObjekte(), sa, role);
              }
              catch (OneSubscriptionPerSendData e)
              {
                // Abmeldung darf nicht von hier erfolgen
                abmeldenErlaubt = false;
              }
              if (abmeldenErlaubt)
              {  
                _sendeAnmeldungen.put(key, sa);
              }  
            }  
          }
        }
        
        ResultData data = new ResultData(resultData.getObject(), sa, resultData.getDataTime(), resultData.getData());
        sendePuffer.add(data);
      }
    }
    
    sendeDaten(sendePuffer);
  }
  
}
