/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package de.bsvrz.vew.sim.sim.steuerung;

import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.Map;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataState;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.vew.sim.sim.EmpfaengerMitAnAbmeldung;

/**
 * Verabeitung von Quittierungen von der SWE StartStopp
 *  
 * @version  $Revision: 1.1 $ / $Date: 2008/10/31 13:30:30 $ / ($Author: Sans $)
 * @author  Dambach-Werke GmbH
 * @author  Stefan Sans
 */
public abstract class StartStopQuittierung implements EmpfaengerMitAnAbmeldung
{
  /**
   * Debug-Logger fr Logging-Ausgaben
   */
  private static final Debug _debug = Debug.getLogger(); 
  
  /**
   * die max. Zeit in ms, die auf das Eintreffen einer Quittierung gewartet wird
   */
  public static final long QUITTUNGS_TIMEOUT = 30000;
  
  private Map<String, String> _posQuittierung;
  private String _prefix;
  private String _suffix;
  private DataDescription _datenBeschreibung;
  private int anzahlQuittierungen;
  private int _warteAufQuittierungen;
  
  /**
   * Bestimmt aus einem Quittierungs-Datensatz, ob die Quittierung positiv ist.
   * @param data der Quittierungs-Datensatz
   * @return true, wenn die Quittierung positiv ist, false sonst
   */
  protected abstract boolean istQuittierungPositiv(Data data);
  
  /**
   * Erzeugt eine neue Instanz
   * 
   * @param prefix Text-Prefix zum Zuordnen einer Start-/Stopp-Antwort
   * @param suffix Text-Suffix zum Zuordnen einer Start-/Stopp-Antwort
   */
  public StartStopQuittierung(String prefix, String suffix)
  {
    _prefix = prefix;
    _suffix = suffix;
    _posQuittierung = new LinkedHashMap<String, String>();
    
    AttributeGroup atg = getConnection().getDataModel().getAttributeGroup("atg.startStoppAntwort");
    Aspect asp = getConnection().getDataModel().getAspect("asp.antwort");
    short simulationsvariante = 0;
    _datenBeschreibung = new DataDescription(atg, asp, simulationsvariante);
  }
  
  /**
   * Bestimme alle positiven Quittierungen
   * @return Read-Only Map mit allen positiven Quittierungen
   */
  public Map<String, String> getPosQuittierung()
  {
    return Collections.unmodifiableMap(_posQuittierung);
  }
  
  /**
   * Bestimme gesamte Anzahl von Quittierungen
   * @return die Anzahl von Quittierungen
   */
  public int getAnzahlQuittierungen()
  {
    return anzahlQuittierungen;
  }
  
  abstract public ClientDavInterface getConnection();
    
  public SystemObject getObjekt()
  {
    return getConnection().getLocalApplicationObject();
  }
  
  public void anmelden()
  {
    getConnection().subscribeReceiver(this, getObjekt(), _datenBeschreibung, 
        ReceiveOptions.normal(), ReceiverRole.receiver());
  }
  
  public void abmelden()
  {
    getConnection().unsubscribeReceiver(this, getObjekt(), _datenBeschreibung);
  }
  
  public void update(ResultData[] resultData)
  {
    for (ResultData data : resultData)
    {  
      _debug.finest("Daten empfangen:\n" + data);
      if (data.getDataState() == DataState.DATA)
      {
        Aspect asp = data.getDataDescription().getAspect();
        AttributeGroup atg = data.getDataDescription().getAttributeGroup();
        short sv = data.getDataDescription().getSimulationVariant();
        
        _debug.finer("Daten empfangen mit " + atg.getPid() + "/" +
            asp.getPid() + "/" + sv  + ":\n" + data);
        
        if (sv == 0 && asp.getPid().equals(_datenBeschreibung.getAspect().getPid()) && 
            atg.getPid().equals(_datenBeschreibung.getAttributeGroup().getPid()))
        {
          _debug.fine("Bearbeite Daten mit " + atg.getPid() + "/" +
              asp.getPid() + "/" + sv + ":\n" + data);
          
          Data datenSatz = data.getData();
          
          int art = datenSatz.getItem("art").asUnscaledValue().intValue();
          int zustand = datenSatz.getItem("zustand").asUnscaledValue().intValue();
          int fehler = datenSatz.getItem("fehler").asUnscaledValue().intValue();
          
          String absenderZeichen = datenSatz.getItem("absenderZeichen").asTextValue().getText();
          String id = datenSatz.getItem("id").asTextValue().getText();
          String artText = datenSatz.getItem("art").asTextValue().getText();
          String zustandText = datenSatz.getItem("zustand").asTextValue().getText();
          String fehlerText = datenSatz.getItem("fehler").asTextValue().getText();
          
          _debug.fine("StartStoppAntwort mit:\n" + 
              "absenderZeichen: " + absenderZeichen + "\n" +
              "id: " + id + "\n" +
              "art: " + artText + "\t" + art + "\n" +
              "zustand: " + zustandText + "\t" + zustand + "\n" +
              "fehler: " + fehlerText + "\t" + fehler + "\n");
          
          if (absenderZeichen.startsWith(_prefix) &&
              absenderZeichen.endsWith(_suffix) &&
              absenderZeichen.indexOf('\\') > 0)
          {
            // Diese Antwort ist fr uns            
            String[] keys = absenderZeichen.split("\\\\");
            if (keys.length > 1 && keys[1].length() > _suffix.length())
            {
              // Diese Antwort ist durch uns verarbeitbar
              String rechner = keys[1].substring(0, keys[1].length() - _suffix.length());
              
              if (istQuittierungPositiv(datenSatz))
              {
                _posQuittierung.put(rechner,id);
              }
              anzahlQuittierungen += 1;
            }
          }
        }
      }
    }
    
    if (_warteAufQuittierungen > 0 &&
        getAnzahlQuittierungen() >= _warteAufQuittierungen)
    {
      synchronized(this)
      {
        notify();
      }
    }
  }
  
  /**
   * Blockiere den aktuellen Thread, bis eine gegebene Anzahl 
   * Quittierungen eingetroffen ist oder der Timeout abgelaufen ist
   * @param quittierungen Anzahl von Quittierungen, die abgewartet werden sollen
   */
  public synchronized void warteAufQuittierungen(int quittierungen)
  {
    _warteAufQuittierungen = quittierungen;
    
    if (_warteAufQuittierungen > getAnzahlQuittierungen())
    { 
      try
      {
        wait(QUITTUNGS_TIMEOUT);
      }
      catch(InterruptedException e)
      {
        
      }
    }
    _warteAufQuittierungen = 0;
  }
  
}

